#include "system.h"
#include "sys/alt_stdio.h"
#include "sys/alt_sys_wrappers.h"
#include "altera_avalon_pio_regs.h"

#include "7SEG.h"

volatile uint8_t displayData[4]={0, 0, 0, 0};

void setDisplay(uint8_t disp){
	switch(disp){  // aktywuj wybrany wyswietlacz
	case DS1:
		IOWR_ALTERA_AVALON_PIO_DATA(DISPLAY_BASE, (1<<DS1));
		break;
	case DS2:
		IOWR_ALTERA_AVALON_PIO_DATA(DISPLAY_BASE, (1<<DS2));
		break;
	case DS3:
		IOWR_ALTERA_AVALON_PIO_DATA(DISPLAY_BASE, (1<<DS3));
		break;
	case DS4:
		IOWR_ALTERA_AVALON_PIO_DATA(DISPLAY_BASE, (1<<DS4));
		break;
	case DSOFF:  // wylacz wszystkie wyswietlacze
	default:
		IOWR_ALTERA_AVALON_PIO_DATA(DISPLAY_BASE, 0);
		break;
	}
}

void refreshDisplay(void){
	static uint8_t disp = 0;

	setDisplay(DSOFF);
	IOWR_ALTERA_AVALON_PIO_DATA(SEGMENT_BASE, displayData[disp]);
	setDisplay(disp);

	if(disp < 3){
		disp++;
	}else{
		disp = 0;
	}
}

// wzory cyfr
const uint8_t  digits[17]={
	(1<<SEG_A)|(1<<SEG_B)|(1<<SEG_C)|(1<<SEG_D)|(1<<SEG_E)|(1<<SEG_F),//0
	(1<<SEG_B)|(1<<SEG_C),//1
	(1<<SEG_A)|(1<<SEG_B)|(1<<SEG_D)|(1<<SEG_E)|(1<<SEG_G),//2
	(1<<SEG_A)|(1<<SEG_B)|(1<<SEG_C)|(1<<SEG_D)|(1<<SEG_G),//3
	(1<<SEG_B)|(1<<SEG_C)|(1<<SEG_F)|(1<<SEG_G),//4
	(1<<SEG_A)|(1<<SEG_C)|(1<<SEG_D)|(1<<SEG_F)|(1<<SEG_G),//5
	(1<<SEG_A)|(1<<SEG_C)|(1<<SEG_D)|(1<<SEG_E)|(1<<SEG_F)|(1<<SEG_G),//6
	(1<<SEG_A)|(1<<SEG_B)|(1<<SEG_C),//7
	(1<<SEG_A)|(1<<SEG_B)|(1<<SEG_C)|(1<<SEG_D)|(1<<SEG_E)|(1<<SEG_F)|(1<<SEG_G),//8
	(1<<SEG_A)|(1<<SEG_B)|(1<<SEG_C)|(1<<SEG_D)|(1<<SEG_F)|(1<<SEG_G),//9
	(1<<SEG_A)|(1<<SEG_B)|(1<<SEG_C)|(1<<SEG_E)|(1<<SEG_F)|(1<<SEG_G),//A
	(1<<SEG_C)|(1<<SEG_D)|(1<<SEG_E)|(1<<SEG_F)|(1<<SEG_G),//B
	(1<<SEG_A)|(1<<SEG_D)|(1<<SEG_E)|(1<<SEG_F),//C
	(1<<SEG_B)|(1<<SEG_C)|(1<<SEG_D)|(1<<SEG_E)|(1<<SEG_G),//D
	(1<<SEG_A)|(1<<SEG_D)|(1<<SEG_E)|(1<<SEG_F)|(1<<SEG_G),//E
	(1<<SEG_A)|(1<<SEG_E)|(1<<SEG_F)|(1<<SEG_G),//F
	(1<<SEG_G)//-
};

void setDigit(uint8_t digit, uint8_t dp, uint8_t pos){
	if(pos < 4){
		if(digit > 16){  // dla wartosci spoza zakresu gasimy wszystkie segmenty
			displayData[pos] = 0|(dp!=0?(1<<SEG_DP):0);
		}else{  // dla pozostalych wartosci wyswietamy odpowiednia liczbe
			displayData[pos] = digits[digit] | (dp!=0?(1<<SEG_DP):0);
		}
	}
}

void intDisplayDec(uint16_t number){
	for(uint8_t i = 0 ; i < 4 ; i++){
		setDigit(number % 10, 0, i);
		number /= 10;
	}
}

void intDisplayDecPoint(uint16_t number, uint8_t pointPosition){
	for(uint8_t i = 0 ; i < 4 ; i++){
		setDigit(number % 10, (i == pointPosition), i);
		number /= 10;
	}
}

void intDisplayDecSpace(uint16_t number){
	for(uint8_t i = 0 ; i < 4 ; i++){
		if(i > 0 && number == 0){
			setDigit(OFF, 0, i);
		}else{
			setDigit(number % 10, 0, i);
		}
		number /= 10;
	}
}

void intDisplayHex(uint16_t number){
	for(uint8_t i = 0 ; i < 4 ; i++){
		setDigit(number % 16, 0, i);
		number /= 16;
	}
}
