/* Includes ------------------------------------------------------------------*/
#include <assert.h>
#include "ring_buffer.h"
#include "sys/alt_irq.h"


bool RingBuffer_Init(RingBuffer *ringBuffer, char *dataBuffer, size_t dataBufferSize) 
{
	assert(ringBuffer);
	assert(dataBuffer);
	assert(dataBufferSize > 0);
	
	if ((ringBuffer) && (dataBuffer) && (dataBufferSize > 0)) {
	  ringBuffer->dataBuffer = dataBuffer;
		ringBuffer->dataBufferSize = dataBufferSize;
		ringBuffer->dataBufferLen = 0;
		ringBuffer->tailIdx = 0;
		
		return true;
	}
	
	return false;
}

bool RingBuffer_Clear(RingBuffer *ringBuffer)
{
	assert(ringBuffer);
	
	if (ringBuffer) {
		ringBuffer->dataBufferLen = 0;
		ringBuffer->tailIdx = 0;
		return true;
	}
	return false;
}

bool RingBuffer_IsEmpty(const RingBuffer *ringBuffer)
{
  assert(ringBuffer);	
	
	if (ringBuffer) {
		return ringBuffer->dataBufferLen == 0;
	}
	
	return true;
	
}

size_t RingBuffer_GetLen(const RingBuffer *ringBuffer)
{
	assert(ringBuffer);
	
	if (ringBuffer) {
		return ringBuffer->dataBufferLen;
	}
	return 0;
	
}

size_t RingBuffer_GetCapacity(const RingBuffer *ringBuffer)
{
	assert(ringBuffer);
	
	if (ringBuffer) {
		return ringBuffer->dataBufferSize;
	}
	return 0;	
}


bool RingBuffer_PutChar(RingBuffer *ringBuffer, char c)
{
	assert(ringBuffer);
	


	if (ringBuffer) {
		alt_irq_context context = alt_irq_disable_all();
		if(ringBuffer->dataBufferLen == ringBuffer->dataBufferSize){
			alt_irq_enable_all(context);
			return false;
		}
		unsigned int writeIndex = (ringBuffer->tailIdx + ringBuffer->dataBufferLen) % ringBuffer->dataBufferSize;
		ringBuffer->dataBuffer[writeIndex] = c;
		ringBuffer->dataBufferLen++;
		alt_irq_enable_all(context);
		return true;
		
	}
	return false;
}

bool RingBuffer_GetChar(RingBuffer *ringBuffer, char *c)
{
	assert(ringBuffer);
	assert(c);
	
  if ((ringBuffer) && (c)) {
	  	alt_irq_context context = alt_irq_disable_all();
		if(ringBuffer->dataBufferLen == 0){
			alt_irq_enable_all(context);
			return false;
		}
		*c = ringBuffer->dataBuffer[ringBuffer->tailIdx];
		ringBuffer->dataBufferLen--;
		ringBuffer->tailIdx = (ringBuffer->tailIdx + 1) % ringBuffer->dataBufferSize;
		alt_irq_enable_all(context);
		return true;
	}
	return false;
}
