#include "i2c.h"

#include "altera_avalon_i2c_regs.h"
#include "altera_avalon_i2c.h"
#include "sys/alt_sys_wrappers.h"

void i2cInit(uint32_t base){
	uint32_t reg = IORD_ALT_AVALON_I2C_CTRL(base);
	reg |= ALT_AVALON_I2C_CTRL_EN_MSK;
	IOWR_ALT_AVALON_I2C_CTRL(base, reg);
}

void i2cDeinit(uint32_t base){
	uint32_t reg = IORD_ALT_AVALON_I2C_CTRL(base);
	reg &= ~ALT_AVALON_I2C_CTRL_EN_MSK;
	IOWR_ALT_AVALON_I2C_CTRL(base, reg);
}

uint8_t i2cSetSpeed(uint32_t base, uint32_t coreHz, uint32_t speedHz){

	uint32_t scl_lcnt,scl_hcnt;

	if ((speedHz > ALT_AVALON_I2C_FS_MAX_HZ) || (speedHz < ALT_AVALON_I2C_SS_MIN_HZ) || (speedHz == 0)){
		return 0;
	}

	// <lcount> = <internal clock> / 2 * <speed, Hz>
	scl_lcnt = (coreHz) / (speedHz << 1);

	// adjust h/l by predetermined amount
	scl_hcnt = scl_lcnt + ALT_AVALON_I2C_DIFF_LCNT_HCNT;
	scl_lcnt = scl_lcnt - ALT_AVALON_I2C_DIFF_LCNT_HCNT;

	if (speedHz > ALT_AVALON_I2C_FS_MIN_HZ){
		uint32_t reg = IORD_ALT_AVALON_I2C_CTRL(base);
		reg |= ALT_AVALON_I2C_CTRL_BUS_SPEED_MSK;
		IOWR_ALT_AVALON_I2C_CTRL(base, reg);
	}else{
		uint32_t reg = IORD_ALT_AVALON_I2C_CTRL(base);
		reg &= ~ALT_AVALON_I2C_CTRL_BUS_SPEED_MSK;
		IOWR_ALT_AVALON_I2C_CTRL(base, reg);
	}

	IOWR_ALT_AVALON_I2C_SCL_HIGH(base, scl_hcnt);
	IOWR_ALT_AVALON_I2C_SCL_LOW(base, scl_lcnt);
	IOWR_ALT_AVALON_I2C_SDA_HOLD(base, scl_lcnt - (scl_lcnt / 2));

	return 1;
}

void i2cWrite(uint32_t base, uint8_t startStop, uint8_t data){
	uint32_t timeout = 10000;
	while(!(IORD_ALT_AVALON_I2C_ISR(base) & ALT_AVALON_I2C_ISR_TX_READY_MSK)){
		if(timeout < 10) ALT_USLEEP(1000);
		if(--timeout == 0) break;
	}
	IOWR_ALT_AVALON_I2C_ISR(base, ALT_AVALON_I2C_ISR_ALL_CLEARABLE_INTS_MSK);

	IOWR_ALT_AVALON_I2C_TFR_CMD(base, (uint32_t)data | ((uint32_t)startStop)<<8);

}

uint8_t i2cRead(uint32_t base, uint8_t startStop){
	i2cWrite(base, startStop, 0);

	uint32_t timeout = 10000;
	while(IORD_ALT_AVALON_I2C_RX_DATA_FIFO_LVL(base) == 0){
		if(timeout < 10) ALT_USLEEP(1000);
		if(--timeout == 0) break;
	}

	return IORD_ALT_AVALON_I2C_RX_DATA(base);
}
